//////////////////////////////////////////////////////////////////////////
// DC6809EMUL.C - Motorola 6809 micropocessor emulation
// Author   : Daniel Coulom - danielcoulom@gmail.com
// Web site : http://dcmo5.free.fr
// Created  : 1997-08
// Last updated : 2008-10-24
//
// This file is part of DCMO5 v11.
// 
// DCMO5 v11 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// DCMO5 v11 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with DCMO5 v11.  If not, see <http://www.gnu.org/licenses/>.
//
//////////////////////////////////////////////////////////////////////////

#include "dcmo5.h"

//global variables
int n;                           //cycle count
int CC = 0x10;                   //condition code
short PC, D, X, Y, U, S, DA, W;  //6809 two byte registers + W work register
char *PCHp, *PCLp, *Ap, *Bp;     //pointers to register bytes
char *XHp, *XLp, *YHp, *YLp;     //pointers to register bytes
char *UHp, *ULp, *SHp, *SLp;     //pointers to register bytes
char *DPp, *DDp, *WHp, *WLp;     //pointers to register bytes

/* memory access C = 1 byte, W = 2 bytes */
#define MGETC(x)   Mgetc(x)
#define MPUTC(x,y) Mputc(x,y)
#define MGETW(x)   (Mgetc(x)<<8|(Mgetc(x+1)&0xff))
#define MPUTW(x,y) {Mputc(x,y>>8);Mputc(x+1,y);}

/*condition code masks (CC=EFHINZVC)*/
#define  CC_C 0x01  /* carry */
#define  CC_V 0x02  /* overflow */
#define  CC_Z 0x04  /* zero */
#define  CC_N 0x08  /* negative */
#define  CC_I 0x10  /* irq mask */
#define  CC_H 0x20  /* half-carry */
#define  CC_F 0x40  /* firq mask */
#define  CC_E 0x80  /* extended registers save */

/*conditional jump*/
/*
C BCC BCS      Z BNE BEQ      V BVC BVS     NZV BGT BLE
0 yes no       0 yes no       0 yes no      000 yes no
1 no  yes      1 no  yes      1 no  yes     100 no  yes
                                            010 no  yes
N BPL BMI     ZC BHI BLS     NV BGE BLT     001 no  yes
0 yes no      00 yes no      00 yes no      110 no  no
1 no  yes     10 no  yes     10 no  yes     101 yes no
              01 no  yes     01 no  yes     011 no  no
              11 no  no      11 yes no      111 no  yes
*/
#define BR_BCC (CC&1)==0  // BCC = BHS
#define BR_BCS (CC&1)==1  // BCS = BLO
#define BR_BVC (CC&2)==0
#define BR_BVS (CC&2)==2
#define BR_BNE (CC&4)==0
#define BR_BEQ (CC&4)==4
#define BR_BHI (CC&5)==0
#define BR_BLS (CC&5)==4||(CC&5)==1
#define BR_BPL (CC&8)==0
#define BR_BMI (CC&8)==8
#define BR_BGE (CC&10)==0||(CC&10)==10
#define BR_BLT (CC&10)==8||(CC&10)==2
#define BR_BGT (CC&14)==0||(CC&14)==10
#define BR_BLE (CC&14)==8||(CC&14)==14||(CC&14)==4||(CC&14)==2
#define BR_BRA {PC+=MGETC(PC);}
#define BR_LBRA {PC+=MGETW(PC);n++;}

//repetitive code
#define X_IND Mgeti()
#define X_DIR *DDp=MGETC(PC);PC++
#define X_EXT W=MGETW(PC);PC+=2
#define SET_Z if(W)CC&=0xfb;else CC|=4

// Init pointers to register bytes ///////////////////////////////////////////
void Init6809registerpointers()
{
 int i = 1;                //integer used to test endianness
 char *c = (char*)&i;      //left byte of integer i  

 PCHp = PCLp = (char*)&PC; //PC msb & lsb
 Ap   = Bp   = (char*)&D;  // D msb & lsb
 XHp  = XLp  = (char*)&X;  // X msb & lsb
 YHp  = YLp  = (char*)&Y;  // Y msb & lsb
 UHp  = ULp  = (char*)&U;  // U msb & lsb
 SHp  = SLp  = (char*)&S;  // S msb & lsb
 DPp  = DDp  = (char*)&DA; //DP msb & lsb
 WHp  = WLp  = (char*)&W;  // W msb & lsb

 switch(c[0]) //test endianness: 1=little-endian, 0=big-endian
 {
  case 1:  PCHp++; Ap++; XHp++; YHp++; UHp++; SHp++; DPp++; WHp++; break;
  default: PCLp++; Bp++; XLp++; YLp++; ULp++; SLp++; DDp++; WLp++; break;
 }                         
}
  
// Get memory (indexed) //////////////////////////////////////////////////////
void Mgeti()
{
 int i;
 short *r;
 i = MGETC(PC); PC++;
 switch (i & 0x60)
 {
  case 0x00: r = &X; break;
  case 0x20: r = &Y; break;
  case 0x40: r = &U; break;
  case 0x60: r = &S; break;
  default: r = &X;
 }
 switch(i &= 0x9f)
 {
  case 0x80: n = 2; W = *r; *r += 1; return;                    // ,R+       
  case 0x81: n = 3; W = *r; *r += 2; return;                    // ,R++      
  case 0x82: n = 2; *r -= 1; W = *r; return;                    // ,-R       
  case 0x83: n = 3; *r -= 2; W = *r; return;                    // ,--R      
  case 0x84: n = 0; W = *r; return;                             // ,R        
  case 0x85: n = 1; W = *r + *Bp; return;                       // B,R       
  case 0x86: n = 1; W = *r + *Ap; return;                       // A,R       
  case 0x87: n = 0; W = *r; return;                             // invalid       
  case 0x88: n = 1; W = *r + MGETC(PC); PC++; return;           // char,R    
  case 0x89: n = 4; X_EXT; W += *r; return;                     // word,R    
  case 0x8a: n = 0; W = *r; return;                             // invalid       
  case 0x8b: n = 4; W = *r + D; return;                         // D,R       
  case 0x8c: n = 1; W = MGETC(PC); PC++; W += PC; return;       // char,PCR  
  case 0x8d: n = 5; X_EXT; W += PC; return;                     // word,PCR  
  case 0x8e: n = 0; W = *r; return;                             // invalid       
  case 0x8f: n = 0; W = *r; return;                             // invalid       
  case 0x90: n = 3; W = MGETW(*r); return;                      // invalid       
  case 0x91: n = 6; *r += 2; W = MGETW(*r - 2); return;         // [,R++]    
  case 0x92: n = 3; W = MGETW(*r); return;                      // invalid       
  case 0x93: n = 6; *r -= 2; W = MGETW(*r); return;             // [,--R]    
  case 0x94: n = 3; W = MGETW(*r); return;                      // [,R]      
  case 0x95: n = 4; W = MGETW(*r + *Bp); return;                // [B,R]     
  case 0x96: n = 4; W = MGETW(*r + *Ap); return;                // [A,R]     
  case 0x97: n = 3; W = MGETW(*r); return;                      // invalid       
  case 0x98: n = 4; W = MGETW(*r + MGETC(PC)); PC++; return;    // [char,R]  
  case 0x99: n = 7; X_EXT; W = MGETW(*r + W); return;           // [word,R]  
  case 0x9a: n = 3; W = MGETW(*r); return;                      // invalid       
  case 0x9b: n = 7; W = MGETW(*r + D); return;                  // [D,R]     
  case 0x9c: n = 4; W = MGETW(PC+1+MGETC(PC)); PC++; return;    // [char,PCR]
  case 0x9d: n = 8; X_EXT; W = MGETW(PC + W); return;           // [word,PCR]
  case 0x9e: n = 3; W = MGETW(*r); return;                      // invalid       
  case 0x9f: n = 5; X_EXT; W = MGETW(W); return;                // [word]    
  default  : n = 1; if(i & 0x10) i -= 0x20; W = *r + i; return; // 5 bits,R  
  //Assumes 0x84 for invalid bytes 0x87 0x8a 0x8e 0x8f
  //Assumes 0x94 for invalid bytes 0x90 0x92 0x97 0x9a 0x9e
 }
}

// PSH, PUL, EXG, TFR /////////////////////////////////////////////////////////
void Pshs(char c)
{
 //Pshs(0xff) = 12 cycles   
 //Pshs(0xfe) = 11 cycles   
 //Pshs(0x80) = 2 cycles   
 if(c & 0x80) {MPUTC(--S,*PCLp); MPUTC(--S,*PCHp); n += 2;}
 if(c & 0x40) {MPUTC(--S, *ULp); MPUTC(--S, *UHp); n += 2;}
 if(c & 0x20) {MPUTC(--S, *YLp); MPUTC(--S, *YHp); n += 2;}
 if(c & 0x10) {MPUTC(--S, *XLp); MPUTC(--S, *XHp); n += 2;}
 if(c & 0x08) {MPUTC(--S, *DPp); n += 1;}
 if(c & 0x04) {MPUTC(--S,  *Bp); n += 1;}
 if(c & 0x02) {MPUTC(--S,  *Ap); n += 1;}
 if(c & 0x01) {MPUTC(--S, CC); n += 1;}
}

void Pshu(char c)
{
 if(c & 0x80) {MPUTC(--U,*PCLp); MPUTC(--U,*PCHp); n += 2;}
 if(c & 0x40) {MPUTC(--U, *SLp); MPUTC(--U, *SHp); n += 2;}
 if(c & 0x20) {MPUTC(--U, *YLp); MPUTC(--U, *YHp); n += 2;}
 if(c & 0x10) {MPUTC(--U, *XLp); MPUTC(--U, *XHp); n += 2;}
 if(c & 0x08) {MPUTC(--U, *DPp); n += 1;}
 if(c & 0x04) {MPUTC(--U,  *Bp); n += 1;}
 if(c & 0x02) {MPUTC(--U,  *Ap); n += 1;}
 if(c & 0x01) {MPUTC(--U, CC); n += 1;}
}

void Puls(char c)
{
 if(c & 0x01) {CC = MGETC(S); S++; n += 1;}
 if(c & 0x02) { *Ap = MGETC(S); S++; n += 1;}
 if(c & 0x04) { *Bp = MGETC(S); S++; n += 1;}
 if(c & 0x08) {*DPp = MGETC(S); S++; n += 1;}
 if(c & 0x10) {*XHp = MGETC(S); S++; *XLp = MGETC(S); S++; n += 2;}
 if(c & 0x20) {*YHp = MGETC(S); S++; *YLp = MGETC(S); S++; n += 2;}
 if(c & 0x40) {*UHp = MGETC(S); S++; *ULp = MGETC(S); S++; n += 2;}
 if(c & 0x80) {*PCHp= MGETC(S); S++; *PCLp= MGETC(S); S++; n += 2;}
}

void Pulu(char c)
{
 if(c & 0x01) {CC = MGETC(U); U++; n += 1;}
 if(c & 0x02) { *Ap = MGETC(U); U++; n += 1;}
 if(c & 0x04) { *Bp = MGETC(U); U++; n += 1;}
 if(c & 0x08) {*DPp = MGETC(U); U++; n += 1;}
 if(c & 0x10) {*XHp = MGETC(U); U++; *XLp = MGETC(U); U++; n += 2;}
 if(c & 0x20) {*YHp = MGETC(U); U++; *YLp = MGETC(U); U++; n += 2;}
 if(c & 0x40) {*SHp = MGETC(U); U++; *SLp = MGETC(U); U++; n += 2;}
 if(c & 0x80) {*PCHp= MGETC(U); U++; *PCLp= MGETC(U); U++; n += 2;}
}

void Exg(char c)
{
 switch(c & 0xff)
 {
  case 0x01: W = D; D = X; X = W; return;    //D-X
  case 0x02: W = D; D = Y; Y = W; return;    //D-Y
  case 0x03: W = D; D = U; U = W; return;    //D-U
  case 0x04: W = D; D = S; S = W; return;    //D-S
  case 0x05: W = D; D = PC; PC = W; return;  //D-PC
  case 0x10: W = X; X = D; D = W; return;    //X-D
  case 0x12: W = X; X = Y; Y = W; return;    //X-Y
  case 0x13: W = X; X = U; U = W; return;    //X-U
  case 0x14: W = X; X = S; S = W; return;    //X-S
  case 0x15: W = X; X = PC; PC = W; return;  //X-PC
  case 0x20: W = Y; Y = D; D = W; return;    //Y-D
  case 0x21: W = Y; Y = X; X = W; return;    //Y-X
  case 0x23: W = Y; Y = U; U = W; return;    //Y-U
  case 0x24: W = Y; Y = S; S = W; return;    //Y-S
  case 0x25: W = Y; Y = PC; PC = W; return;  //Y-PC
  case 0x30: W = U; U = D; D = W; return;    //U-D
  case 0x31: W = U; U = X; X = W; return;    //U-X
  case 0x32: W = U; U = Y; Y = W; return;    //U-Y
  case 0x34: W = U; U = S; S = W; return;    //U-S
  case 0x35: W = U; U = PC; PC = W; return;  //U-PC
  case 0x40: W = S; S = D; D = W; return; 
  case 0x41: W = S; S = X; X = W; return; 
  case 0x42: W = S; S = Y; Y = W; return; 
  case 0x43: W = S; S = U; U = W; return; 
  case 0x45: W = S; S = PC; PC = W; return; 
  case 0x50: W = PC; PC = D; D = W; return; 
  case 0x51: W = PC; PC = X; X = W; return; 
  case 0x52: W = PC; PC = Y; Y = W; return; 
  case 0x53: W = PC; PC = U; U = W; return; 
  case 0x54: W = PC; PC = S; S = W; return; 
  case 0x89: W = *Ap; *Ap = *Bp; *Bp = W; return; 
  case 0x8a: W = *Ap; *Ap = CC; CC = W; return; 
  case 0x8b: W = *Ap; *Ap = *DPp; *DPp = W; return; 
  case 0x98: W = *Bp; *Bp = *Ap; *Ap = W; return; 
  case 0x9a: W = *Bp; *Bp = CC; CC = W; return; 
  case 0x9b: W = *Bp; *Bp = *DPp; *DPp = W; return; 
  case 0xa8: W = CC; CC = *Ap; *Ap = W; return; 
  case 0xa9: W = CC; CC = *Bp; *Bp = W; return; 
  case 0xab: W = CC; CC = *DPp; *DPp = W; return; 
  case 0xb8: W = *DPp; *DPp = *Ap; *Ap = W; return; 
  case 0xb9: W = *DPp; *DPp = *Bp; *Bp = W; return; 
  case 0xba: W = *DPp; *DPp = CC; CC = W; return; 
 }
}

void Tfr(char c)
{
 switch(c & 0xff)
 {
  case 0x01: X = D; return; 
  case 0x02: Y = D; return; 
  case 0x03: U = D; return; 
  case 0x04: S = D; return; 
  case 0x05: PC = D; return; 
  case 0x10: D = X; return; 
  case 0x12: Y = X; return; 
  case 0x13: U = X; return; 
  case 0x14: S = X; return; 
  case 0x15: PC = X; return; 
  case 0x20: D = Y; return; 
  case 0x21: X = Y; return; 
  case 0x23: U = Y; return; 
  case 0x24: S = Y; return; 
  case 0x25: PC = Y; return; 
  case 0x30: D = U; return; 
  case 0x31: X = U; return; 
  case 0x32: Y = U; return; 
  case 0x34: S = U; return; 
  case 0x35: PC = U; return; 
  case 0x40: D = S; return; 
  case 0x41: X = S; return; 
  case 0x42: Y = S; return; 
  case 0x43: U = S; return; 
  case 0x45: PC = S; return; 
  case 0x50: D = PC; return; 
  case 0x51: X = PC; return; 
  case 0x52: Y = PC; return; 
  case 0x53: U = PC; return; 
  case 0x54: S = PC; return; 
  case 0x89: *Bp = *Ap; return; 
  case 0x8a: CC = *Ap; return; 
  case 0x8b: *DPp = *Ap; return; 
  case 0x98: *Ap = *Bp; return; 
  case 0x9a: CC = *Bp; return; 
  case 0x9b: *DPp = *Bp; return; 
  case 0xa8: *Ap = CC; return; 
  case 0xa9: *Bp = CC; return; 
  case 0xab: *DPp = CC; return; 
  case 0xb8: *Ap = *DPp; return; 
  case 0xb9: *Bp = *DPp; return; 
  case 0xba: CC = *DPp; return; 
 }
}

// CLR, NEG, COM, INC, DEC  (CC=EFHINZVC) /////////////////////////////////////
char Clr()
{
 CC &= 0xf0;
 CC |= CC_Z;
 return 0;
}

char Neg(char c)
{
 CC &= 0xf0;
 if(c == -128) CC |= CC_V;
 c = - c;
 if(c != 0) CC |= CC_C;
 if(c < 0) CC |= CC_N;
 if(c == 0) CC |= CC_Z; 
 return c;
}

char Com(char c)
{
 CC &= 0xf0;
 c = ~c;
 CC |= CC_C;
 if(c < 0) CC |= CC_N;
 if(c == 0) CC |= CC_Z;
 return c;
}

char Inc(char c)
{
 CC &= 0xf1;
 if(c == 127) CC |= CC_V;
 c++;
 if(c < 0) CC |= CC_N;
 if(c == 0) CC |= CC_Z;
 return c;
}

char Dec(char c)
{
 CC &= 0xf1;
 if(c == -128) CC |= CC_V;
 c--;
 if(c < 0) CC |= CC_N;
 if(c == 0) CC |= CC_Z;
 return c;
}

// Registers operations  (CC=EFHINZVC) ////////////////////////////////////////
void Mul()
{
 D = (*Ap & 0xff) * (*Bp & 0xff);
 CC &= 0xf2;
 if(D < 0) CC |= CC_C;
 if(D == 0) CC |= CC_Z;
}

void Addc(char *r, char c)
{
 int i = *r + c;
 CC &= 0xd0;
 if(((*r & 0x0f) + (c & 0x0f)) & 0x10) CC |= CC_H;
 if(((*r & 0xff) + (c & 0xff)) & 0x100) CC |= CC_C;
 *r = i & 0xff;
 if(*r != i) CC |= CC_V;
 if(*r < 0) CC |= CC_N;
 if(*r == 0) CC |= CC_Z;
}

void Adc(char *r, char c)
{
 int carry = (CC & CC_C);
 int i = *r + c + carry;
 CC &= 0xd0;
 if(((*r & 0x0f) + (c & 0x0f) + carry) & 0x10) CC |= CC_H;
 if(((*r & 0xff) + (c & 0xff) + carry) & 0x100) CC |= CC_C;
 *r = i & 0xff;
 if(*r != i) CC |= CC_V;
 if(*r < 0) CC |= CC_N;
 if(*r == 0) CC |= CC_Z;
}

void Addw(short *r, short w)
{
 int i = *r + w;
 CC &= 0xf0;
 if(((*r & 0xffff) + (w & 0xffff)) & 0xf0000) CC |= CC_C;
 *r = i & 0xffff;
 if(*r != i) CC |= CC_V;
 if(*r < 0) CC |= CC_N;
 if(*r == 0) CC |= CC_Z;
}

void Subc(char *r, char c)
{
 int i = *r - c;
 CC &= 0xf0;
 if(((*r & 0xff) - (c & 0xff)) & 0x100) CC |= CC_C;
 *r = i & 0xff;
 if(*r != i) CC |= CC_V;
 if(*r < 0) CC |= CC_N;
 if(*r == 0) CC |= CC_Z;
}

void Sbc(char *r, char c)
{
 int carry = (CC & CC_C);
 int i = *r - c - carry;
 CC &= 0xf0;
 if(((*r & 0xff) - (c & 0xff) - carry) & 0x100) CC |= CC_C;
 *r = i & 0xff;
 if(*r != i) CC |= CC_V;
 if(*r < 0) CC |= CC_N;
 if(*r == 0) CC |= CC_Z;
}

void Subw(short *r, short w)
{
 int i = *r - w;
 CC &= 0xf0;
 if(((*r & 0xffff) - (w & 0xffff)) & 0x10000) CC |= CC_C;
 *r = i & 0xffff;
 if(*r != i) CC |= CC_V;
 if(*r < 0) CC |= CC_N;
 if(*r == 0) CC |= CC_Z;
}

void Daa()
{
 int i = *Ap & 0xff;
 if((CC & CC_H) || ((i & 0x00f) > 0x09)) i += 0x06;
 if((CC & CC_C) || ((i & 0x1f0) > 0x90)) i += 0x60;
 *Ap = i & 0xff;
 i = (i >> 1 & 0xff) | (CC << 7);
 CC &= 0xf0;
 if(i & 0x80) CC |= CC_C;
 if((*Ap ^ i) & 0x80) CC |= CC_V;
 if(*Ap < 0) CC |= CC_N;
 if(*Ap == 0) CC |= CC_Z;
}

// Shift and rotate  (CC=EFHINZVC) ////////////////////////////////////////////
char Lsr(char c)
{
 CC &= 0xf2;
 if(c & 1) CC |= CC_C;
 c = (c & 0xff) >> 1; 
 if(c == 0) CC |= CC_Z;
 return c;
}

char Ror(char c)
{
 int carry = CC & CC_C; 
 CC &= 0xf2;
 if(c & 1) CC |= CC_C;
 c = ((c & 0xff) >> 1) | (carry << 7);
 if(c < 0) CC |= CC_N;
 if(c == 0) CC |= CC_Z;
 return c;
}

char Rol(char c)
{
 int carry = CC & CC_C;
 CC &= 0xf0;
 if(c < 0) CC |= CC_C;
 c = ((c & 0x7f) << 1) | carry;
 if((c >> 7 & 1) ^ (CC & CC_C)) CC |= CC_V;
 if(c < 0) CC |= CC_N;
 if(c == 0) CC |= CC_Z;
 return c;
}

char Asr(char c)
{
 CC &= 0xf2;
 if(c & 1) CC |= CC_C;
 c = ((c & 0xff) >> 1) | (c & 0x80);
 if(c < 0) CC |= CC_N;
 if(c == 0) CC |= CC_Z;
 return c;
}

char Asl(char c)
{
 CC &= 0xf0;
 if(c < 0) CC |= CC_C;
 c = (c & 0xff) << 1;
 if((c >> 7 & 1) ^ (CC & CC_C)) CC |= CC_V;
 if(c < 0) CC |= CC_N;
 if(c == 0) CC |= CC_Z;
 return c;
}

// Test and compare  (CC=EFHINZVC) ////////////////////////////////////////////
void Tstc(char c)
{
 CC &= 0xf1;
 if(c < 0) CC |= CC_N;
 if(c == 0) CC |= CC_Z;
}

void Tstw(short w)
{
 CC &= 0xf1;
 if(w < 0) CC |= CC_N;
 if(w == 0) CC |= CC_Z;
}

void Cmpc(char *reg, char c)
{
 char r = *reg;
 int i = *reg - c;
 CC &= 0xf0;
 if(((r & 0xff) - (c & 0xff)) & 0x100) CC |= CC_C;
 r = i & 0xff;
 if(r != i) CC |= CC_V;
 if(r < 0) CC |= CC_N;
 if(r == 0) CC |= CC_Z;
}

void Cmpw(short *reg, short w)
{
 short r = *reg;
 int i = *reg - w;
 CC &= 0xf0;
 if(((r & 0xffff) - (w & 0xffff)) & 0x10000) CC |= CC_C;
 r = i & 0xffff;
 if(r != i) CC |= CC_V;
 if(r < 0) CC |= CC_N;
 if(r == 0) CC |= CC_Z;
}

// Interrupt requests  (CC=EFHINZVC) //////////////////////////////////////////
void Swi(int n)
{
 CC |= CC_E;
 Pshs(0xff);
 CC |= CC_I;
 if(n == 1) {PC = MGETW(0xfffa); return;}
 if(n == 2) {PC = MGETW(0xfff4); return;}
 if(n == 3) {PC = MGETW(0xfff2); return;}
} 

void Irq()
{
 if((CC & CC_I) == 0)
 {CC |= CC_E; Pshs(0xff); CC |= CC_I; PC = MGETW(0xfff8);}
}

void Firq()
{
 if((CC & CC_F) == 0)
 {CC &= ~CC_E; Pshs(0x81); CC |= CC_F; CC |= CC_I; PC = MGETW(0xfff6);}
}

// RTI ////////////////////////////////////////////////////////////////////////
void Rti() {Puls(0x01); if(CC & CC_E) Puls(0xfe); else Puls(0x80);}

// Execute one operation at PC address and set PC to next opcode address //////
int Run6809()
/*
Return value is set to :
- cycle count for the executed instruction when operation code is legal
- negative value (-code) when operation code is illegal
*/
{
 int precode, code;
 n = 0; //par defaut pas de cycles supplementaires
 precode = 0; //par defaut pas de precode
 while(1)
 {   
  code = MGETC(PC) & 0xff; PC++;
  if(code == 0x10) {precode = 0x1000; continue;}           
  if(code == 0x11) {precode = 0x1100; continue;}           
  code |= precode; break;
 }

 switch(code)
 {
  case 0x00: X_DIR; MPUTC(DA, Neg(MGETC(DA))); return 6;      /* NEG  /$ */
  case 0x01: X_DIR; return 3;                            /*undoc BRN     */
//case 0x02: 
// if(CC&C1){MPUTC(wd, Com(MGETC(DA))); return 6;}       /*undoc COM  /$ */
//      else{MPUTC(DA, Neg(MGETC(DA))); return 6;}       /*undoc NEG  /$ */
  case 0x03: X_DIR; MPUTC(DA, Com(MGETC(DA))); return 6;      /* COM  /$ */
  case 0x04: X_DIR; MPUTC(DA, Lsr(MGETC(DA))); return 6;      /* LSR  /$ */
//case 0x05: MPUTC(DA, Lsr(MGETC(DA))); return 6;        /*undoc LSR  /$ */
  case 0x06: X_DIR; MPUTC(DA, Ror(MGETC(DA))); return 6;      /* ROR  /$ */
  case 0x07: X_DIR; MPUTC(DA, Asr(MGETC(DA))); return 6;      /* ASR  /$ */
  case 0x08: X_DIR; MPUTC(DA, Asl(MGETC(DA))); return 6;      /* ASL  /$ */
  case 0x09: X_DIR; MPUTC(DA, Rol(MGETC(DA))); return 6;      /* ROL  /$ */
  case 0x0a: X_DIR; MPUTC(DA, Dec(MGETC(DA))); return 6;      /* DEC  /$ */
  case 0x0c: X_DIR; MPUTC(DA, Inc(MGETC(DA))); return 6;      /* INC  /$ */
  case 0x0d: X_DIR; Tstc(MGETC(DA)); return 6;                /* TST  /$ */
  case 0x0e: X_DIR; PC = DA; return 3;                        /* JMP  /$ */
  case 0x0f: X_DIR; MPUTC(DA, Clr()); return 6;               /* CLR  /$ */

  case 0x12: return 2;                                        /* NOP     */
  case 0x13: return 4;                                        /* SYNC    */
  case 0x16: PC += MGETW(PC) + 2; return 5;                   /* LBRA    */
  case 0x17: X_EXT; Pshs(0x80); PC += W; return 9;            /* LBSR    */
  case 0x19: Daa(); return 2;                                 /* DAA     */
  case 0x1a: CC |= MGETC(PC); PC++; return 3;                 /* ORCC #$ */
  case 0x1c: CC &= MGETC(PC); PC++; return 3;                 /* ANDC #$ */
  case 0x1d: Tstw(D = *Bp); return 2;                         /* SEX     */
  case 0x1e: PC++; Exg(MGETC(PC-1)); return 8;                /* EXG     */
  case 0x1f: PC++; Tfr(MGETC(PC-1)); return 6;                /* TFR     */

  case 0x20: BR_BRA; PC++; return 3;                          /* BRA     */
  case 0x21: PC++; return 3;                                  /* BRN     */
  case 0x22: if(BR_BHI) BR_BRA; PC++; return 3;               /* BHI     */
  case 0x23: if(BR_BLS) BR_BRA; PC++; return 3;               /* BLS     */
  case 0x24: if(BR_BCC) BR_BRA; PC++; return 3;               /* BCC     */
  case 0x25: if(BR_BCS) BR_BRA; PC++; return 3;               /* BCS     */
  case 0x26: if(BR_BNE) BR_BRA; PC++; return 3;               /* BNE     */
  case 0x27: if(BR_BEQ) BR_BRA; PC++; return 3;               /* BEQ     */
  case 0x28: if(BR_BVC) BR_BRA; PC++; return 3;               /* BVC     */
  case 0x29: if(BR_BVS) BR_BRA; PC++; return 3;               /* BVS     */
  case 0x2a: if(BR_BPL) BR_BRA; PC++; return 3;               /* BPL     */
  case 0x2b: if(BR_BMI) BR_BRA; PC++; return 3;               /* BMI     */
  case 0x2c: if(BR_BGE) BR_BRA; PC++; return 3;               /* BGE     */
  case 0x2d: if(BR_BLT) BR_BRA; PC++; return 3;               /* BLT     */
  case 0x2e: if(BR_BGT) BR_BRA; PC++; return 3;               /* BGT     */
  case 0x2f: if(BR_BLE) BR_BRA; PC++; return 3;               /* BLE     */

  case 0x30: X_IND; X = W; SET_Z; return 4 + n;               /* LEAX    */
  case 0x31: X_IND; Y = W; SET_Z; return 4 + n;               /* LEAY    */
  //d'apres Prehisto, LEAX et LEAY positionnent aussi le bit N de CC
  //il faut donc modifier l'mulation de ces deux instructions
  case 0x32: X_IND; S = W; return 4 + n; /*CC not set*/       /* LEAS    */
  case 0x33: X_IND; U = W; return 4 + n; /*CC not set*/       /* LEAU    */
  case 0x34: PC++; Pshs(MGETC(PC-1)); return 5 + n;           /* PSHS    */
  case 0x35: PC++; Puls(MGETC(PC-1)); return 5 + n;           /* PULS    */
  case 0x36: PC++; Pshu(MGETC(PC-1)); return 5 + n;           /* PSHU    */
  case 0x37: PC++; Pulu(MGETC(PC-1)); return 5 + n;           /* PULU    */
  case 0x39: Puls(0x80); return 5;                            /* RTS     */
  case 0x3a: X += *Bp & 0xff; return 3;                       /* ABX     */
  case 0x3b: Rti(); return 4 + n;                             /* RTI     */
  case 0x3c: CC &= MGETC(PC); PC++; CC |= CC_E; return 20;    /* CWAI    */
  case 0x3d: Mul(); return 11;                                /* MUL     */
  case 0x3f: Swi(1); return 19;                               /* SWI     */

  case 0x40: *Ap = Neg(*Ap); return 2;                        /* NEGA    */
  case 0x43: *Ap = Com(*Ap); return 2;                        /* COMA    */
  case 0x44: *Ap = Lsr(*Ap); return 2;                        /* LSRA    */
  case 0x46: *Ap = Ror(*Ap); return 2;                        /* RORA    */
  case 0x47: *Ap = Asr(*Ap); return 2;                        /* ASRA    */
  case 0x48: *Ap = Asl(*Ap); return 2;                        /* ASLA    */
  case 0x49: *Ap = Rol(*Ap); return 2;                        /* ROLA    */
  case 0x4a: *Ap = Dec(*Ap); return 2;                        /* DECA    */
  case 0x4c: *Ap = Inc(*Ap); return 2;                        /* INCA    */
  case 0x4d: Tstc(*Ap); return 2;                             /* TSTA    */
  case 0x4f: *Ap = Clr(); return 2;                           /* CLRA    */

  case 0x50: *Bp = Neg(*Bp); return 2;                        /* NEGB    */
  case 0x53: *Bp = Com(*Bp); return 2;                        /* COMB    */
  case 0x54: *Bp = Lsr(*Bp); return 2;                        /* LSRB    */
  case 0x56: *Bp = Ror(*Bp); return 2;                        /* RORB    */
  case 0x57: *Bp = Asr(*Bp); return 2;                        /* ASRB    */
  case 0x58: *Bp = Asl(*Bp); return 2;                        /* ASLB    */
  case 0x59: *Bp = Rol(*Bp); return 2;                        /* ROLB    */
  case 0x5a: *Bp = Dec(*Bp); return 2;                        /* DECB    */
  case 0x5c: *Bp = Inc(*Bp); return 2;                        /* INCB    */
  case 0x5d: Tstc(*Bp); return 2;                             /* TSTB    */
  case 0x5f: *Bp = Clr(); return 2;                           /* CLRB    */

  case 0x60: X_IND; MPUTC(W, Neg(MGETC(W))); return 6 + n;    /* NEG  IX */
  case 0x63: X_IND; MPUTC(W, Com(MGETC(W))); return 6 + n;    /* COM  IX */
  case 0x64: X_IND; MPUTC(W, Lsr(MGETC(W))); return 6 + n;    /* LSR  IX */
  case 0x66: X_IND; MPUTC(W, Ror(MGETC(W))); return 6 + n;    /* ROR  IX */
  case 0x67: X_IND; MPUTC(W, Asr(MGETC(W))); return 6 + n;    /* ASR  IX */
  case 0x68: X_IND; MPUTC(W, Asl(MGETC(W))); return 6 + n;    /* ASL  IX */
  case 0x69: X_IND; MPUTC(W, Rol(MGETC(W))); return 6 + n;    /* ROL  IX */
  case 0x6a: X_IND; MPUTC(W, Dec(MGETC(W))); return 6 + n;    /* DEC  IX */
  case 0x6c: X_IND; MPUTC(W, Inc(MGETC(W))); return 6 + n;    /* INC  IX */
  case 0x6d: X_IND; Tstc(MGETC(W)); return 6 + n;             /* TST  IX */
  case 0x6e: X_IND; PC = W; return 3 + n;                     /* JMP  IX */
  case 0x6f: X_IND; MPUTC(W, Clr()); return 6 + n;            /* CLR  IX */

  case 0x70: X_EXT; MPUTC(W, Neg(MGETC(W))); return 7;        /* NEG  $  */
  case 0x73: X_EXT; MPUTC(W, Com(MGETC(W))); return 7;        /* COM  $  */
  case 0x74: X_EXT; MPUTC(W, Lsr(MGETC(W))); return 7;        /* LSR  $  */
  case 0x76: X_EXT; MPUTC(W, Ror(MGETC(W))); return 7;        /* ROR  $  */
  case 0x77: X_EXT; MPUTC(W, Asr(MGETC(W))); return 7;        /* ASR  $  */
  case 0x78: X_EXT; MPUTC(W, Asl(MGETC(W))); return 7;        /* ASL  $  */
  case 0x79: X_EXT; MPUTC(W, Rol(MGETC(W))); return 7;        /* ROL  $  */
  case 0x7a: X_EXT; MPUTC(W, Dec(MGETC(W))); return 7;        /* DEC  $  */
  case 0x7c: X_EXT; MPUTC(W, Inc(MGETC(W))); return 7;        /* INC  $  */
  case 0x7d: X_EXT; Tstc(MGETC(W)); return 7;                 /* TST  $  */
  case 0x7e: X_EXT; PC = W; return 4;                         /* JMP  $  */
  case 0x7f: X_EXT; MPUTC(W, Clr()); return 7;                /* CLR  $  */

  case 0x80: Subc(Ap, MGETC(PC)); PC++; return 2;             /* SUBA #$ */
  case 0x81: Cmpc(Ap, MGETC(PC)); PC++; return 2;             /* CMPA #$ */
  case 0x82: Sbc(Ap, MGETC(PC)); PC++; return 2;              /* SBCA #$ */
  case 0x83: X_EXT; Subw(&D, W); return 4;                    /* SUBD #$ */
  case 0x84: Tstc(*Ap &= MGETC(PC)); PC++; return 2;          /* ANDA #$ */
  case 0x85: Tstc(*Ap & MGETC(PC)); PC++; return 2;           /* BITA #$ */
  case 0x86: Tstc(*Ap = MGETC(PC)); PC++; return 2;           /* LDA  #$ */
  case 0x88: Tstc(*Ap ^= MGETC(PC)); PC++; return 2;          /* EORA #$ */
  case 0x89: Adc(Ap, MGETC(PC)); PC++; return 2;              /* ADCA #$ */
  case 0x8a: Tstc(*Ap |= MGETC(PC)); PC++; return 2;          /* ORA  #$ */
  case 0x8b: Addc(Ap, MGETC(PC)); PC++; return 2;             /* ADDA #$ */
  case 0x8c: X_EXT; Cmpw(&X, W); return 4;                    /* CMPX #$ */
  case 0x8d: X_DIR; Pshs(0x80); PC += *DDp; return 7;         /* BSR     */
  case 0x8e: X_EXT; Tstw(X = W); return 3;                    /* LDX  #$ */

  case 0x90: X_DIR; Subc(Ap, MGETC(DA)); return 4;            /* SUBA /$ */
  case 0x91: X_DIR; Cmpc(Ap, MGETC(DA)); return 4;            /* CMPA /$ */
  case 0x92: X_DIR; Sbc(Ap, MGETC(DA)); return 4;             /* SBCA /$ */
  case 0x93: X_DIR; Subw(&D, MGETW(DA));return 6;             /* SUBD /$ */
  case 0x94: X_DIR; Tstc(*Ap &= MGETC(DA)); return 4;         /* ANDA /$ */
  case 0x95: X_DIR; Tstc(*Ap & MGETC(DA)); return 4;          /* BITA /$ */
  case 0x96: X_DIR; Tstc(*Ap = MGETC(DA)); return 4;          /* LDA  /$ */
  case 0x97: X_DIR; MPUTC(DA, *Ap); Tstc(*Ap); return 4;      /* STA  /$ */
  case 0x98: X_DIR; Tstc(*Ap ^= MGETC(DA)); return 4;         /* EORA /$ */
  case 0x99: X_DIR; Adc(Ap, MGETC(DA)); return 4;             /* ADCA /$ */
  case 0x9a: X_DIR; Tstc(*Ap |= MGETC(DA)); return 4;         /* ORA  /$ */
  case 0x9b: X_DIR; Addc(Ap, MGETC(DA)); return 4;            /* ADDA /$ */
  case 0x9c: X_DIR; Cmpw(&X, MGETW(DA)); return 6;            /* CMPX /$ */
  case 0x9d: X_DIR; Pshs(0x80); PC = DA; return 7;            /* JSR  /$ */
  case 0x9e: X_DIR; Tstw(X = MGETW(DA)); return 5;            /* LDX  /$ */
  case 0x9f: X_DIR; MPUTW(DA, X); Tstw(X); return 5;          /* STX  /$ */

  case 0xa0: X_IND; Subc(Ap, MGETC(W)); return 4 + n;         /* SUBA IX */
  case 0xa1: X_IND; Cmpc(Ap, MGETC(W)); return 4 + n;         /* CMPA IX */
  case 0xa2: X_IND; Sbc(Ap, MGETC(W)); return 4 + n;          /* SBCA IX */
  case 0xa3: X_IND; Subw(&D, MGETW(W)); return 6 + n;         /* SUBD IX */
  case 0xa4: X_IND; Tstc(*Ap &= MGETC(W)); return 4 + n;      /* ANDA IX */
  case 0xa5: X_IND; Tstc(MGETC(W) & *Ap); return 4 + n;       /* BITA IX */
  case 0xa6: X_IND; Tstc(*Ap = MGETC(W)); return 4 + n;       /* LDA  IX */
  case 0xa7: X_IND; MPUTC(W,*Ap); Tstc(*Ap); return 4 + n;    /* STA  IX */
  case 0xa8: X_IND; Tstc(*Ap ^= MGETC(W)); return 4 + n;      /* EORA IX */
  case 0xa9: X_IND; Adc(Ap, MGETC(W)); return 4 + n;          /* ADCA IX */
  case 0xaa: X_IND; Tstc(*Ap |= MGETC(W)); return 4 + n;      /* ORA  IX */
  case 0xab: X_IND; Addc(Ap, MGETC(W)); return 4 + n;         /* ADDA IX */
  case 0xac: X_IND; Cmpw(&X, MGETW(W)); return 4 + n;         /* CMPX IX */
  case 0xad: X_IND; Pshs(0x80); PC = W; return 5 + n;         /* JSR  IX */
  case 0xae: X_IND; Tstw(X = MGETW(W)); return 5 + n;         /* LDX  IX */
  case 0xaf: X_IND; MPUTW(W, X); Tstw(X); return 5 + n;       /* STX  IX */

  case 0xb0: X_EXT; Subc(Ap, MGETC(W)); return 5;             /* SUBA $  */
  case 0xb1: X_EXT; Cmpc(Ap, MGETC(W)); return 5;             /* CMPA $  */
  case 0xb2: X_EXT; Sbc(Ap, MGETC(W)); return 5;              /* SBCA $  */
  case 0xb3: X_EXT; Subw(&D, MGETW(W)); return 7;             /* SUBD $  */
  case 0xb4: X_EXT; Tstc(*Ap &= MGETC(W)); return 5;          /* ANDA $  */
  case 0xb5: X_EXT; Tstc(*Ap & MGETC(W)); return 5;           /* BITA $  */
  case 0xb6: X_EXT; Tstc(*Ap = MGETC(W)); return 5;           /* LDA  $  */
  case 0xb7: X_EXT; MPUTC(W, *Ap); Tstc(*Ap); return 5;       /* STA  $  */
  case 0xb8: X_EXT; Tstc(*Ap ^= MGETC(W)); return 5;          /* EORA $  */
  case 0xb9: X_EXT; Adc(Ap, MGETC(W)); return 5;              /* ADCA $  */
  case 0xba: X_EXT; Tstc(*Ap |= MGETC(W)); return 5;          /* ORA  $  */
  case 0xbb: X_EXT; Addc(Ap, MGETC(W)); return 5;             /* ADDA $  */
  case 0xbc: X_EXT; Cmpw(&X, MGETW(W)); return 7;             /* CMPX $  */
  case 0xbd: X_EXT; Pshs(0x80); PC = W; return 8;             /* JSR  $  */
  case 0xbe: X_EXT; Tstw(X = MGETW(W)); return 6;             /* LDX  $  */
  case 0xbf: X_EXT; MPUTW(W, X); Tstw(X); return 6;           /* STX  $  */

  case 0xc0: Subc(Bp, MGETC(PC)); PC++; return 2;             /* SUBB #$ */
  case 0xc1: Cmpc(Bp, MGETC(PC)); PC++; return 2;             /* CMPB #$ */
  case 0xc2: Sbc(Bp, MGETC(PC)); PC++; return 2;              /* SBCB #$ */
  case 0xc3: X_EXT; Addw(&D, W); return 4;                    /* ADDD #$ */
  case 0xc4: Tstc(*Bp &= MGETC(PC)); PC++; return 2;          /* ANDB #$ */
  case 0xc5: Tstc(*Bp & MGETC(PC)); PC++; return 2;           /* BITB #$ */
  case 0xc6: Tstc(*Bp = MGETC(PC)); PC++; return 2;           /* LDB  #$ */
  case 0xc8: Tstc(*Bp ^= MGETC(PC)); PC++; return 2;          /* EORB #$ */
  case 0xc9: Adc(Bp, MGETC(PC)); PC++; return 2;              /* ADCB #$ */
  case 0xca: Tstc(*Bp |= MGETC(PC)); PC++; return 2;          /* ORB  #$ */
  case 0xcb: Addc(Bp, MGETC(PC)); PC++;return 2;              /* ADDB #$ */
  case 0xcc: X_EXT; Tstw(D = W); return 3;                    /* LDD  #$ */
  case 0xce: X_EXT; Tstw(U = W); return 3;                    /* LDU  #$ */

  case 0xd0: X_DIR; Subc(Bp, MGETC(DA)); return 4;            /* SUBB /$ */
  case 0xd1: X_DIR; Cmpc(Bp, MGETC(DA)); return 4;            /* CMPB /$ */
  case 0xd2: X_DIR; Sbc(Bp, MGETC(DA)); return 4;             /* SBCB /$ */
  case 0xd3: X_DIR; Addw(&D, MGETW(DA)); return 6;            /* ADDD /$ */
  case 0xd4: X_DIR; Tstc(*Bp &= MGETC(DA)); return 4;         /* ANDB /$ */
  case 0xd5: X_DIR; Tstc(MGETC(DA) & *Bp); return 4;          /* BITB /$ */
  case 0xd6: X_DIR; Tstc(*Bp = MGETC(DA)); return 4;          /* LDB  /$ */
  case 0xd7: X_DIR; MPUTC(DA,*Bp); Tstc(*Bp); return 4;       /* STB  /$ */
  case 0xd8: X_DIR; Tstc(*Bp ^= MGETC(DA)); return 4;         /* EORB /$ */
  case 0xd9: X_DIR; Adc(Bp, MGETC(DA)); return 4;             /* ADCB /$ */
  case 0xda: X_DIR; Tstc(*Bp |= MGETC(DA)); return 4;         /* ORB  /$ */
  case 0xdb: X_DIR; Addc(Bp, MGETC(DA)); return 4;            /* ADDB /$ */
  case 0xdc: X_DIR; Tstw(D = MGETW(DA)); return 5;            /* LDD  /$ */
  case 0xdd: X_DIR; MPUTW(DA, D); Tstw(D); return 5;          /* STD  /$ */
  case 0xde: X_DIR; Tstw(U = MGETW(DA)); return 5;            /* LDU  /$ */
  case 0xdf: X_DIR; MPUTW(DA, U); Tstw(U); return 5;          /* STU  /$ */

  case 0xe0: X_IND; Subc(Bp, MGETC(W)); return 4 + n;         /* SUBB IX */
  case 0xe1: X_IND; Cmpc(Bp, MGETC(W)); return 4 + n;         /* CMPB IX */
  case 0xe2: X_IND; Sbc(Bp, MGETC(W)); return 4 + n;          /* SBCB IX */
  case 0xe3: X_IND; Addw(&D, MGETW(W)); return 6 + n;         /* ADDD IX */
  case 0xe4: X_IND; Tstc(*Bp &= MGETC(W)); return 4 + n;      /* ANDB IX */
  case 0xe5: X_IND; Tstc(*Bp & MGETC(W)); return 4 + n;       /* BITB IX */
  case 0xe6: X_IND; Tstc(*Bp = MGETC(W)); return 4 + n;       /* LDB  IX */
  case 0xe7: X_IND; MPUTC(W, *Bp); Tstc(*Bp); return 4 + n;   /* STB  IX */
  case 0xe8: X_IND; Tstc(*Bp ^= MGETC(W)); return 4 + n;      /* EORB IX */
  case 0xe9: X_IND; Adc(Bp, MGETC(W)); return 4 + n;          /* ADCB IX */
  case 0xea: X_IND; Tstc(*Bp |= MGETC(W)); return 4 + n;      /* ORB  IX */
  case 0xeb: X_IND; Addc(Bp, MGETC(W)); return 4 + n;         /* ADDB IX */
  case 0xec: X_IND; Tstw(D = MGETW(W)); return 5 + n;         /* LDD  IX */
  case 0xed: X_IND; MPUTW(W, D); Tstw(D); return 5 + n;       /* STD  IX */
  case 0xee: X_IND; Tstw(U = MGETW(W)); return 5 + n;         /* LDU  IX */
  case 0xef: X_IND; MPUTW(W, U); Tstw(U); return 5 + n;       /* STU  IX */

  case 0xf0: X_EXT; Subc(Bp, MGETC(W)); return 5;             /* SUBB $  */
  case 0xf1: X_EXT; Cmpc(Bp, MGETC(W)); return 5;             /* CMPB $  */
  case 0xf2: X_EXT; Sbc(Bp, MGETC(W)); return 5;              /* SBCB $  */
  case 0xf3: X_EXT; Addw(&D, MGETW(W)); return 7;             /* ADDD $  */
  case 0xf4: X_EXT; Tstc(*Bp &= MGETC(W)); return 5;          /* ANDB $  */
  case 0xf5: X_EXT; Tstc(*Bp & MGETC(W)); return 5;           /* BITB $  */
  case 0xf6: X_EXT; Tstc(*Bp = MGETC(W)); return 5;           /* LDB  $  */
  case 0xf7: X_EXT; MPUTC(W, *Bp); Tstc(*Bp); return 5;       /* STB  $  */
  case 0xf8: X_EXT; Tstc(*Bp ^= MGETC(W)); return 5;          /* EORB $  */
  case 0xf9: X_EXT; Adc(Bp, MGETC(W)); return 5;              /* ADCB $  */
  case 0xfa: X_EXT; Tstc(*Bp |= MGETC(W)); return 5;          /* ORB  $  */
  case 0xfb: X_EXT; Addc(Bp, MGETC(W)); return 5;             /* ADDB $  */
  case 0xfc: X_EXT; Tstw(D = MGETW(W)); return 6;             /* LDD  $  */
  case 0xfd: X_EXT; MPUTW(W, D); Tstw(D); return 6;           /* STD  $  */
  case 0xfe: X_EXT; Tstw(U = MGETW(W)); return 6;             /* LDU  $  */
  case 0xff: X_EXT; MPUTW(W, U); Tstw(U); return 6;           /* STU  $  */

  case 0x1021: PC += 2; return 5;                             /* LBRN    */
  case 0x1022: if(BR_BHI) BR_LBRA; PC += 2; return 5 + n;     /* LBHI    */
  case 0x1023: if(BR_BLS) BR_LBRA; PC += 2; return 5 + n;     /* LBLS    */
  case 0x1024: if(BR_BCC) BR_LBRA; PC += 2; return 5 + n;     /* LBCC    */
  case 0x1025: if(BR_BCS) BR_LBRA; PC += 2; return 5 + n;     /* LBCS    */
  case 0x1026: if(BR_BNE) BR_LBRA; PC += 2; return 5 + n;     /* LBNE    */
  case 0x1027: if(BR_BEQ) BR_LBRA; PC += 2; return 5 + n;     /* LBEQ    */
  case 0x1028: if(BR_BVC) BR_LBRA; PC += 2; return 5 + n;     /* LBVC    */
  case 0x1029: if(BR_BVS) BR_LBRA; PC += 2; return 5 + n;     /* LBVS    */
  case 0x102a: if(BR_BPL) BR_LBRA; PC += 2; return 5 + n;     /* LBPL    */
  case 0x102b: if(BR_BMI) BR_LBRA; PC += 2; return 5 + n;     /* LBMI    */
  case 0x102c: if(BR_BGE) BR_LBRA; PC += 2; return 5 + n;     /* LBGE    */
  case 0x102d: if(BR_BLT) BR_LBRA; PC += 2; return 5 + n;     /* LBLT    */
  case 0x102e: if(BR_BGT) BR_LBRA; PC += 2; return 5 + n;     /* LBGT    */
  case 0x102f: if(BR_BLE) BR_LBRA; PC += 2; return 5 + n;     /* LBLE    */
  case 0x103f: Swi(2); return 20;                             /* SWI2    */

  case 0x1083: X_EXT; Cmpw(&D, W); return 5;                  /* CMPD #$ */
  case 0x108c: X_EXT; Cmpw(&Y, W); return 5;                  /* CMPY #$ */
  case 0x108e: X_EXT; Tstw(Y = W); return 4;                  /* LDY  #$ */
  case 0x1093: X_DIR; Cmpw(&D, MGETW(DA)); return 7;          /* CMPD /$ */
  case 0x109c: X_DIR; Cmpw(&Y, MGETW(DA)); return 7;          /* CMPY /$ */
  case 0x109e: X_DIR; Tstw(Y = MGETW(DA)); return 6;          /* LDY  /$ */
  case 0x109f: X_DIR; MPUTW(DA, Y); Tstw(Y); return 6;        /* STY  /$ */
  case 0x10a3: X_IND; Cmpw(&D, MGETW(W)); return 7 + n;       /* CMPD IX */
  case 0x10ac: X_IND; Cmpw(&Y, MGETW(W)); return 7 + n;       /* CMPY IX */
  case 0x10ae: X_IND; Tstw(Y = MGETW(W)); return 6 + n;       /* LDY  IX */
  case 0x10af: X_IND; MPUTW(W, Y); Tstw(Y); return 6 + n;     /* STY  IX */
  case 0x10b3: X_EXT; Cmpw(&D, MGETW(W)); return 8;           /* CMPD $  */
  case 0x10bc: X_EXT; Cmpw(&Y, MGETW(W)); return 8;           /* CMPY $  */
  case 0x10be: X_EXT; Tstw(Y = MGETW(W)); return 7;           /* LDY  $  */
  case 0x10bf: X_EXT; MPUTW(W, Y); Tstw(Y); return 7;         /* STY  $  */
  case 0x10ce: X_EXT; Tstw(S = W); return 4;                  /* LDS  #$ */
  case 0x10de: X_DIR; Tstw(S = MGETW(DA)); return 6;          /* LDS  /$ */
  case 0x10df: X_DIR; MPUTW(DA, S); Tstw(S); return 6;        /* STS  /$ */
  case 0x10ee: X_IND; Tstw(S = MGETW(W)); return 6 + n;       /* LDS  IX */
  case 0x10ef: X_IND; MPUTW(W, S); Tstw(S); return 6 + n;     /* STS  IX */
  case 0x10fe: X_EXT; Tstw(S = MGETW(W)); return 7;           /* LDS  $  */
  case 0x10ff: X_EXT; MPUTW(W, S); Tstw(S); return 7;         /* STS  $  */

  case 0x113f: Swi(3); return 20;                             /* SWI3    */
  case 0x1183: X_EXT; Cmpw(&U, W); return 5;                  /* CMPU #$ */
  case 0x118c: X_EXT; Cmpw(&S, W); return 5;                  /* CMPS #$ */
  case 0x1193: X_DIR; Cmpw(&U, MGETW(DA)); return 7;          /* CMPU /$ */
  case 0x119c: X_DIR; Cmpw(&S, MGETW(DA)); return 7;          /* CMPS /$ */
  case 0x11a3: X_IND; Cmpw(&U, MGETW(W)); return 7 + n;       /* CMPU IX */
  case 0x11ac: X_IND; Cmpw(&S, MGETW(W)); return 7 + n;       /* CMPS IX */
  case 0x11b3: X_EXT; Cmpw(&U, MGETW(W)); return 8;           /* CMPU $  */
  case 0x11bc: X_EXT; Cmpw(&S, MGETW(W)); return 8;           /* CMPS $  */

  default: return -code;                                      /* Illegal */
 } 
}
